/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	debug potential memory leak (or other)	*/
/*	trouble.				*/
/*						*/
/************************************************/
#include	<malloc.h>
#include	<errno.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<time.h>
#include	<syslog.h>
#include	<execinfo.h>
#include	<sys/time.h>
#include	<memory.h>
#include	"dbgmem.h"

#undef	calloc
#undef	malloc
#undef	realloc
#undef	free
#undef	setenv
#undef	unsetenv
#undef	strdup
#undef	asprintf
#undef	vasprintf


#define MXTB	10	/*max trace deep		*/

extern char **environ;	/*application env varirable	*/


typedef	struct timeval	TIMEVAL;

typedef struct leaktyp LEAKTYP;
struct leaktyp	{	/*leak detection structure	*/
	TIMEVAL date;	/*allocation date		*/
	size_t size;	/*memory size			*/
	void *tb[MXTB]; /*back trace			*/
	int noleak;	/*No leak consideration flag	*/
	void *ptr;	/*allocated memory		*/
	LEAKTYP *prev;	/*previous allocated memory	*/
	LEAKTYP *next;	/*next allocated memory		*/
	};

int memleak=DMFALSE;	/*flag to check for memory leak	*/

static LEAKTYP *memory;	/*allocated memory structure	*/
static long nummem;	/*number of leak structure	*/
static char *cpath;	/*Application curpath		*/
static char *cbase;	/*Application base		*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information, from the backtrace stand	*/
/*	point.					*/
/*						*/
/************************************************/
static int sametrace(const LEAKTYP *l1,const LEAKTYP *l2)

{
int inx;
long diff;

inx=0;
diff=0;
while ((inx<MXTB)&&(diff==(long)0)) {
  diff=(long)(l1->tb[inx])-(long)(l2->tb[inx]);
  inx++;
  }
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information				*/
/*						*/
/************************************************/
static int cmpleak(const void *p1,const void *p2)

{
int diff;
LEAKTYP *l1;
LEAKTYP *l2;

l1=((LEAKTYP **)p1)[0];
l2=((LEAKTYP **)p2)[0];
if ((diff=sametrace(l1,l2))==0)
  diff=l1->date.tv_sec-l2->date.tv_sec;
  if (diff==0)
    diff=l1->date.tv_usec-l2->date.tv_usec;
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert back trace address	*/
/*	to sources line information.		*/
/*						*/
/************************************************/
static void disline(LEAKTYP *leak,const char *path,FILE *fichier)

{
#define ADDR2LINE	"addr2line"

int i;
FILE *canal;
char command[1000];

(void) sprintf(command,"%s -e %s",ADDR2LINE,path);
for (i=0;i<MXTB;i++) {
  char str[20];

  if (leak->tb[i]==0)
    break;
  (void) sprintf(str," %08lx",(long)leak->tb[i]);
  (void) strcat(command,str);
  }
if ((canal=popen(command,"r"))!=(FILE *)0) {
  char line[200];

  while (fgets(line,sizeof(line)-1,canal)!=(char *)0) {
    (void) fprintf(fichier,"%s",line);
    }
  (void) fclose(canal);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne a Backtrace stack	*/
/*	(to trace back leak) to memory structure*/
/*						*/
/************************************************/
static void setbtrace(LEAKTYP *leak,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i,j;
struct timezone tz;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (j=0,i=step+1;(j<MXTB)&&(i<num);i++,j++) {
  leak->tb[j]=zone[i];
  }
(void) gettimeofday(&(leak->date),&tz);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report backtrace via syslog*/
/*						*/
/************************************************/
static void logbacktrace(int loglev,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (i=step;i<num;i++) {
  (void) syslog(loglev,"\tbt--> %08lx",(long)zone[i]);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to locate a 'leak memory'	*/
/*	among the list of memory.		*/
/*						*/
/************************************************/
static void *findmemory(register void *ptr)

{
register LEAKTYP *scan;

scan=memory;
while (scan!=(LEAKTYP *)0) {
  if (scan->ptr==ptr) {
    break;
    }
  scan=scan->next;
  }
return scan;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get a "leak proof" memory	*/
/*	email to a shell script			*/
/*						*/
/************************************************/
static void *getmemory(size_t size,int step)

{
register LEAKTYP *leak;

leak=(LEAKTYP *)calloc(1,sizeof(LEAKTYP));
leak->size=size;
leak->ptr=malloc(size);
leak->next=memory;
(void) setbtrace(leak,step+1);
if (memory!=(LEAKTYP *)0)
  memory->prev=leak;
memory=leak;
nummem++;
return leak->ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application path, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void dbg_setcurpath(char *path)

{
cpath=strdup(path);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application base, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void dbg_setcurbase(char *path)

{
cbase=strdup(path);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne new memory area	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_malloc(size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(size,1);
  }
else
  ptr=malloc(size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne clean new memory 	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_calloc(size_t nmemb,size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(nmemb*size,1);
  (void) memset(ptr,'\000',nmemb*size);
  }
else
  ptr=calloc(nmemb,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory from memry pool*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void dbg_free(void *ptr)

{
if (memleak==DMTRUE) {
  LEAKTYP *leak;

  if ((leak=findmemory(ptr))!=(LEAKTYP *)0) {
    if (leak->prev!=(LEAKTYP *)0) {
      leak->prev->next=leak->next;
      }
    else {
      memory=leak->next;
      if (memory!=(LEAKTYP *)0)
        memory->prev=(LEAKTYP *)0;
      }
    if (leak->next!=(LEAKTYP *)0)
      leak->next->prev=leak->prev;
    (void) free(leak->ptr);
    (void) free(leak);
    nummem--;
    }
  else {
    char *strloc="subsys.c:dbg_free, Unable to find 'leak memory' "
		 "for ptr '%08lx'=<%s>\n";

    (void) syslog(LOG_INFO,strloc,(long)ptr,(char *)ptr);
    leak=(LEAKTYP *)calloc(1,sizeof(LEAKTYP));
    (void) setbtrace(leak,0);
    (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
    (void) disline(leak,cpath,stderr);
    (void) fprintf(stderr,"\n");
    (void) free(leak);
    }
  }
else
  (void) free(ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a variable from the	*/
/*	environement list			*/
/*						*/
/************************************************/
int dbg_unsetenv(const char *name)

{
int status;

status=-1;
errno=EINVAL;
if (memleak==DMTRUE) {
  if (environ!=(char **)0) {
    int i;

    for (i=0;environ[i]!=(char *)0;i++) {
      char *ptr;

      if ((ptr=strstr(environ[i],name))!=environ[i])
	continue;
      if (*(environ[i]+strlen(name))=='=') {
	(void) dbg_free(ptr);
        do {
          environ[i]=environ[i+1];
	  i++;
	  }
        while (environ[i]!=(char *)0);
	status=0;
	break;
	}
      }
    }
  }
else 
  status=unsetenv(name);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set variable within the	*/
/*	environement				*/
/*						*/
/************************************************/
int dbg_setenv(const char *name,const char *value,int overwrite)

{
int status;

status=0;
if (memleak==DMTRUE) {
  char *newvar;
  char *ptr;

  (void) dbg_asprintf(&newvar,"%s=%s",name,value);
  if ((ptr=getenv(name))!=(char *)0) {
    if (overwrite!=0) 
      (void) dbg_unsetenv(name);
    else {
      (void) dbg_free(newvar);
      newvar=(char *)0;
      }
    }
  if (newvar!=(char *)0) {
    int taille;

    taille=0;
    if (environ!=(char **)0)
      for (;environ[taille]!=(char *)0;taille++);
    else 
      environ=dbg_calloc(1,sizeof(char *));
    environ=(char **)dbg_realloc((void *)environ,(taille+2)*sizeof(char *));
    environ[taille]=newvar;
    environ[taille+1]=(char *)0;
    }
  }
else {
  status=setenv(name,value,overwrite);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory from pole*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_realloc(void *ptr,size_t size)

{
if (memleak==DMTRUE) {
  if (ptr==(void *)0)
    ptr=getmemory(size,1);
  else {
    LEAKTYP *leak;

    if ((leak=findmemory(ptr))!=(LEAKTYP *)0) {
      leak->ptr=realloc(leak->ptr,size);
      ptr=leak->ptr;
      (void) setbtrace(leak,1);
      }
    else {
      char *strloc="subsys.c:dbg_realloc, Unable to find 'leak memory' "
		   "for ptr '%08lx'=<%s>\n";

      (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
      (void) syslog(LOG_DEBUG,strloc,(long)ptr,(char *)ptr);
      (void) logbacktrace(LOG_DEBUG,2);
      ptr=(void *)0;
      }
    }
  }
else
  ptr=realloc(ptr,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string.			*/
/*						*/
/************************************************/
char *dbg_strdup(const char *str)

{
char *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(strlen(str)+3,1);
  (void) strcpy(ptr,str);
  }
else
  ptr=strdup(str);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory using a va_list	*/
/*						*/
/************************************************/
int dbg_vasprintf(char **strp, const char *fmt,va_list args)

{
int count;

count=-1;
if (memleak==DMTRUE) {
  char *strloc;

  count=vasprintf(&strloc,fmt,args);
  *strp=dbg_strdup(strloc);
  (void) free(strloc);
  }
else {
  count=vasprintf(strp,fmt,args);
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory			*/
/*						*/
/************************************************/
int dbg_asprintf(char **strp, const char *fmt, ...)

{
int count;
va_list args;

count=-1;
va_start(args,fmt);
if (memleak==DMTRUE)
  count=dbg_vasprintf(strp,fmt,args);
else 
  count=vasprintf(strp,fmt,args);
va_end(args);
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory allocated by	*/
/*	a library procedure (as scandir)	*/
/*	Memory allocated by such library routine*/
/*	are not known by leak detection.	*/
/*						*/
/************************************************/
void dbg_sysfree(void *ptr)

{
(void) free(ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to flag current memory as not	*/
/*	to be checked against memory leak	*/
/*	Implementation is such memory can be	*/
/*	flagged/unflagged as a no memory leak.	*/
/*	noleak value should always be positive.	*/
/*						*/
/************************************************/
void dbg_setnoleak(register int value)

{
register LEAKTYP *scan;

scan=memory;
while (scan!=(LEAKTYP *)0) {
  scan->noleak+=value;
  scan=scan->next;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump current memory	*/
/*	allocation and detect memory allocation	*/
/*						*/
/*	Data will be dumped on /tmp under	*/
/*	application name and process pid.	*/
/*						*/
/************************************************/
void dbg_dumpmem(const char *apname,const char *extension)

{
#define	DUMPFRM	"%s/var/tmp/%s%s%s.%05d"

if (memleak==DMTRUE) {
  char *sep;
  register int dummycnt;
  FILE *fichier;
  long collected;
  char *name;
  LEAKTYP **memsort;
  LEAKTYP *leak;

  sep="-";
  collected=0;
  if ((extension==(char *)0)||(strlen(extension)==0)) {
    sep="";
    extension="";
    }
  dummycnt=asprintf(&name,DUMPFRM,cbase,apname,sep,extension,getpid());
  if ((fichier=fopen(name,"w"))==(FILE *)0) {
    (void) fprintf(stderr,"dbg_dumpmem unable to open file <%s> (error=<%s>)\n",
			  name,strerror(errno));
    fichier=stderr;
    }
  memsort=(LEAKTYP **)0;
  if (nummem>0) {
    memsort=(LEAKTYP **)calloc(nummem,sizeof(LEAKTYP *)); 
    leak=memory;
    while (leak!=(LEAKTYP *)0) {
      if (leak->noleak==0) {
        memsort[collected]=leak;
        collected++;
	}
      leak=leak->next;
      }
    }
  if (collected>0) {
    register int i;

    (void) fprintf(fichier,"%05ld memory allocation still open\n",collected);
    (void) qsort(memsort,collected,sizeof(LEAKTYP *),cmpleak);
    leak=(LEAKTYP *)0;
    for (i=0;i<collected;i++) {
      #define STRTIME	"%Y-%m-%d %H:%M:%S"
      struct tm *tm;
      char strtime[40];

      if ((leak==(LEAKTYP *)0)||(sametrace(leak,memsort[i])!=0)) {
        (void) fprintf(fichier,"\n");
	leak=memsort[i];
	(void) disline(leak,cpath,fichier);
	}
      tm=localtime(&(memsort[i]->date.tv_sec));
      (void) strftime(strtime,sizeof(strtime),STRTIME,tm);
      (void) fprintf(fichier,"\t size=%05ld age='%s.%06ld' ptr=%08lx->'%s'\n",
		     	     (long)(memsort[i]->size),
			     strtime,memsort[i]->date.tv_usec,
			     (long)(memsort[i]->ptr),
			     (char *)(memsort[i]->ptr));
      }
    }
  else {
#ifdef DEBUGMEM
    (void) fprintf(fichier,"No memory leak detected\n");
#else
    (void) fprintf(fichier,"No memory leak detector\n(comment in "
			   "'#define DEBUGMEM' within dbgmem.h)\n");
#endif
    }
  (void) fclose(fichier);
  (void) free(name);
  }
}
