/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	LVLEML:					*/
/*	Management level routines to handle	*/
/*	SMTP Protocol exchange with remote site.*/
/*						*/
/************************************************/
#include	<stdarg.h>
#include	<malloc.h>
#include	<errno.h>
#include	<ctype.h>
#include	<string.h>
#include	<syslog.h>
#include	<netdb.h>
#include	<unistd.h>
#include	<arpa/inet.h>
#include	<sys/utsname.h>
#include	"subrou.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"uniper.h"
#include	"gescnf.h"
#include	"gesque.h"
#include	"lvleml.h"

#define MXTIME	300	/*wait for a 300 sec max*/

typedef enum	{
	m_subject,	/*subject header line	*/
	m_mime,		/*MIME version hdr line	*/
	m_contype,	/*MIME contents type	*/
	m_unknown	/*unknown header info	*/
	}MENUM;

typedef	struct	{	/*mime structure	*/
	char *spmrk;	/*spam string marker	*/
	char *mimevers;	/*Mime Version		*/
	char *boundery;	/*mime bondery marker	*/
	char *contype;	/*Mime default contype	*/
	char *subject;	/*header subject info	*/
	}MIMTYP;

typedef	enum	{	/*SMTP vocabulary code	*/
	c_help,		/*provide help		*/
	c_mail,		/*mail from		*/
	c_rcpt,		/*recipient		*/
	c_ehlo,		/*code extension	*/
	c_helo,		/*code presentation	*/
	c_data,		/*data mode		*/
	c_tls,		/*Transport Socket Layer*/
	c_auth,		/*user authentication	*/
	c_reset,	/*reset current stat	*/ 
	c_noop,		/*operation NOOP	*/
	c_quit,		/*abort connection	*/
	c_chkid,	/*user checking ID	*/
	c_ckid,		/*to get site dfltcode	*/
	c_goodmail,	/*it was a good mail	*/
	c_copy,		/*copy file to directory*/
	c_orgn,		/*origin override	*/
	c_unknown	/*unknown code		*/
	}CODTYP;

static	char *hlphelo[]={
		"HELO <hostname>",
		"\tIntroduce yourself.",
		(char *)0
		};

static	char *hlpehlo[]={
		"EHLO <hostname>",
		"\tIntroduce yourself and request extended SMTP mode.",
		"\t8BITMIME\tUse 8-bit data\t\t[RFC1652]",
		"\tSIZE\tMessage size declaration\t\t[RFC1870]",
		"\tSTARTTLS\tSecure SMTP\t\t[RFC2487]",
		(char *)0
		};

static	char *hlpmail[]={
		"MAIL FROM: <sender> [ <parameters> ]",
		"\tSpecifies the sender.  Parameters are ESMTP extensions.",
		(char *)0
		};

static	char *hlprcpt[]={
		"RCPT TO: <recipient> [ <parameters> ]",
		"\tSpecifies the recipient. Can be used any number of times.",
		"\tParameters are ESMTP extensions.",
		(char *)0
		};

static	char *hlpdata[]={
		"DATA",
		"\tFollowing text is collected as the message.",
		"\tEnd with a single dot.",
		(char *)0
		};

static	char *hlpreset[]={
		"RSET",
		"\tResets the system.",
		(char *)0
		};

static	char *hlpnoop[]={
		"NOOP",
		"\tDo nothing.",
		(char *)0
		};

static	char *hlpquit[]={
		"QUIT",
		"\tExit sendmail (SMTP).",
		(char *)0
		};

static	char *hlphelp[]={
		"HELP [ <topic> ]",
		"\tThe HELP command gives help info.",
		(char *)0
		};

static	char *hlptls[]={
		"STARTTLS",
		"\tStart TLS negotiation.",
		(char *)0
		};

static	char *hlpauth[]={
		"AUTH mechanism [initial-response]",
		"\tStart authentication.",
		(char *)0
		};

static	char *hlpchkid[]={
		"CHECKID <userid> password uniqid",
		"\tto check a user id and provide further "
		"access to clement WEB interface",
		(char *)0
		};

static	char *hlpckid[]={
		"CKID dfltcode",
		"\tsend the remote system authentication",
		(char *)0
		};

static	char *hlpgoodmail[]={
		GMAIL " goodmail recipient e_mail_uniq_id",
		"\tto send information about a mail being good.",
		(char *)0
		};

static	char *hlpcopy[]={
		"COPY filename directory",
		"\tCopy some file name to the specific directory.",
		"\tReturn an acknoledge status",
		(char *)0
		};

static	char *hlporgn[]={
		"ORGN <ipnumber>",
		"\tOverride the remote detection (debugging only)",
		(char *)0
		};

typedef	struct	{	/*available command	*/
	CODTYP code;	/*test code		*/	
	SCODTYP scode;	/*related smtp code	*/	
	int proceed;	/*continue to proceed	*/
	char **help;	/*command explainations	*/
	}ACTTYP;

static ACTTYP action[]={ /*MUST in CODTYP order	*/
		{c_help,s_unknown,true,hlphelp},
		{c_mail,s_unknown,true,hlpmail},
		{c_rcpt,s_rcpt,false,hlprcpt},
		{c_ehlo,s_unknown,true,hlpehlo},
		{c_helo,s_unknown,true,hlphelo},
		{c_data,s_data,false,hlpdata},
		{c_tls,s_unknown,true,hlptls},
		{c_auth,s_auth,false,hlpauth},
		{c_reset,s_reset,false,hlpreset},
		{c_noop,s_unknown,true,hlpnoop},
		{c_quit,s_quit,false,hlpquit},
		{c_chkid,s_chkid,false,hlpchkid},
		{c_ckid,s_id,false,hlpckid},
		{c_goodmail,s_goodmail,false,hlpgoodmail},
		{c_copy,s_unknown,true,hlpcopy},
		{c_orgn,s_unknown,true,hlporgn},
		{c_unknown,s_unknown,true,(char **)0}
		};

static VOCTYP vocsmtp[]={
		{c_helo,"HELO",(const void *)(action+c_helo)},
		{c_ehlo,"EHLO",(const void *)(action+c_ehlo)},
		{c_mail,MAILF,(const void *)(action+c_mail)},
		{c_rcpt,RCPT,(const void *)(action+c_rcpt)},
		{c_data,"DATA",(const void *)(action+c_data)},
		{c_tls,"STARTTLS",(const void *)(action+c_tls)},
		{c_auth,"AUTH",(const void *)(action+c_auth)},
		{c_reset,"RSET",(const void *)(action+c_reset)},
		{c_noop,"NOOP",(const void *)(action+c_noop)},
		{c_quit,SQUIT,(const void *)(action+c_quit)},
		{c_help,"HELP",(const void *)(action+c_help)},
		{c_orgn,"ORGN",(const void *)(action+c_orgn)},
		{c_unknown,(char *)0,(const void *)0}
		};

static const VOCTYP vocchkid[]={
		{c_chkid,"CHECKID FOR",(const void *)(action+c_chkid)},
		{c_ckid,CKID,(const void *)(action+c_ckid)},
		{c_goodmail,GMAIL,(const void *)(action+c_goodmail)},
		{c_copy,COPY,(const void *)(action+c_copy)},
		{c_tls,"STARTTLS",(const void *)(action+c_tls)},
		{c_quit,SQUIT,(const void *)(action+c_quit)},
		{c_help,"HELP",(const void *)(action+c_help)},
		{c_unknown,(char *)0,(const void *)0}
		};
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the keyword list	*/
/*	according the protocol used.		*/
/*						*/
/************************************************/
static const VOCTYP *getkwlst(PROTYP proto)

{
const VOCTYP *list;

list=vocsmtp;
switch(proto) {
  case pro_smtp		:
    break;
  case pro_chkid	:
    list=vocchkid;
    break;
  default		:
    (void) alert(0,"lvleml.c:getkwlst, Unexpected prototyp '%d'",proto);
  }
return list;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch data coming from	*/
/*	SMTP stream, spliting line according	*/
/*	'\r'.					*/
/*						*/
/************************************************/
static char **getstream(CTTYP *cta)

{
char **data;
char *buffer;

data=(char **)0;
if ((buffer=tls_waitforaction(MXTIME,cta->tls))!=(char *)0) {
  register char *ptr;
  register char *split;

  ptr=buffer;
  while ((split=strchr(ptr,'\r'))!=(char *)0) {
    if (*(split+1)=='\n') {
      *split='\n';
      split++;
      }
    *split='\000';
    split++;
    data=(char **)rou_addveclst((void **)data,(void *)strdup(ptr)); 
    ptr=split;
    }
  if (strlen(ptr)>0) {
    data=(char **)rou_addveclst((void **)data,(void *)strdup(ptr)); 
    }
  buffer=rou_freestr(buffer);
  }
return data;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove any memory used by	*/
/*	an from user information structure	*/
/*						*/
/************************************************/
static FRMTYP *cleanfrm(FRMTYP *frm)

{
if (frm!=(FRMTYP *)0) {
  if (frm->from!=(char *)0)
    (void) free(frm->from);
  if (frm->listenip!=(char *)0)
    (void) free(frm->listenip);
  if (frm->ipremote!=(char *)0)
    (void) free(frm->ipremote);
  (void) free(frm);
  frm=(FRMTYP *)0;
  }
return frm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure remove any memory used by 	*/
/*	a reciptiens information list		*/
/*						*/
/************************************************/
static RCPTYP **cleanrcpts(RCPTYP **rcpts)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    rcpts[i]=eml_freercpt(rcpts[i]);
    }
  (void) free(rcpts);
  rcpts=(RCPTYP **)0;
  }
return rcpts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the comment line	*/
/*	to help user to decide if it is spam or	*/
/*	not.					*/
/*						*/
/************************************************/
static char *getspamresend(char *webgood,char *emailid,
			   char *spamperti,char *rcpt)

{
char *resend;
int lang;
int cal;
int perti;

resend=(char *)0;
lang=0;
cal=0;
perti=0;
if (sscanf(spamperti,"%d/%d/%d",&lang,&cal,&perti)==3) {
  char *ptr;

  ptr=webgood;
  if (ptr==(char *)0) {
    struct utsname buname;

    if (uname(&buname)==0) {
      if ((ptr=strchr(buname.nodename,'.'))!=(char *)0) 
        ptr++;
      }
    }
  if ((ptr!=(char *)0)&&(strlen(ptr)>0)) {
    const char *rsd;
    register int dummycnt;

    rsd=per_getstrurl((LNGTYP)lang,u_mailgood);
    dummycnt=asprintf(&resend,rsd,ptr,rcpt,emailid,per_getstrlng(lang));
    }
  else {
    (void) alert(0,"%s Unable to find domainname (error=<%s>)",
                   "lvleml.c:getspamresend,",strerror(errno));
    }
  }
return resend;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan an Email and set	*/
/*	spammarker inside.			*/
/*						*/
/************************************************/
static void spammark(MLTYP *ml,char *webgood,char *queudir,char *emailid,
		     char *spamperti,char *rcpt)

{
#define	OPSP	"lvleml.c:spammark,"

register int dummycnt;
char *resend;
char *newname;
char *marker;
FILE *fout;
int cat;
int done;
int phase;
int proceed;
char ename[300];

dummycnt=0;
resend=(char *)0;
newname=(char *)0;
marker=(char *)0;
fout=(FILE *)0;
cat=0;
done=false;
phase=0;
proceed=true;
(void) rou_mkname(ename,sizeof(ename),d_spool,queudir,emailid,EMAILEXT);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting a file name	*/
      if ((resend=getspamresend(webgood,emailid,spamperti,rcpt))==(char *)0) {
	(void) alert(0,"%s Unable to extract spam mark from <%s>",
			OPSP,spamperti);
	proceed=false;
	}
      break;
    case 1	:	/*setting a file name	*/
      dummycnt=asprintf(&marker,"--%s",emailid);
      dummycnt=asprintf(&newname,"%s.tmp",ename);
      if ((fout=fopen(newname,"w"))==(FILE *)0) {
	(void) alert(0,"%s Unable to mark spaminfo within Email <%s> (error=<%s>)",
			OPSP,newname,strerror(errno));
	phase=999;	/*get out fast		*/
	}
      break;
    case 2	:	/*calculating SPAM lvl	*/
      cat=2;		/*hard coded for now	*/
      break;
    case 3	:	/*setting header	*/
      (void) fprintf(fout,"From: <%s>\n",ml->from);
      (void) fprintf(fout,"Date: %s\n",rou_ascsystime(time((time_t *)0)));
      (void) fprintf(fout,"Message-Id: <%s-QUAR>\n",emailid);
      (void) fprintf(fout,"Subject: [SPAM cat %d] <%s>\n",cat,ml->subject);
      (void) fprintf(fout,"%sID: <%s>\n",XCLEM,emailid);
      (void) fprintf(fout,"MIME-Version: 1.0\n");
      (void) fprintf(fout,"Content-Type: multipart/mixed;\n\tboundary=\"%s\"\n",marker);
      (void) fprintf(fout,"\nThis is a multi-part message in MIME format.\n\n");
      break;
    case 4	:	/*adding spam comment	*/
      (void) fprintf(fout,"--%s\n",marker);
      (void) fprintf(fout,"Content-Type: text/html; charset=UTF-8\n");
      (void) fprintf(fout,"Content-Transfer-Encoding: 8bit\n\n");
      (void) fprintf(fout,"%s\n\n",resend);
      break;
    case 5	:	/*flag mime attachment	*/
      (void) fprintf(fout,"--%s\n",marker);
      (void) fprintf(fout,"Content-Type: message/rfc822;\n\tname=\"%s\"\n",ml->subject);
      (void) fprintf(fout,"Content-Transfer-Encoding: 8bit\n");
      (void) fprintf(fout,"Content-Disposition: attachment;\n\tfilename=\"%s\"\n\n",ml->subject);
      break;
    case  6   :       /*appening file         */
      done=rou_appenfile(fout,ename);
      break;
    case  7   :       /*End marker and close	*/
      (void) fprintf(fout,"\n--%s--\n\n",marker);
      (void) fclose(fout);
      break;
    case  8   :       /*moving file around	*/
      if (done==true) {
        (void) unlink(ename);
        if (link(newname,ename)<0)
	  (void) alert(0,"%s Unable to link file name <%s> to <%s> (bug?)",
			  OPSP,newname,ename);
	}
      (void) unlink(newname);
      break;
    default	:	/*SAFE guard		*/
      (void) free(newname);
      (void) free(marker);
      (void) free(resend);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a reply to a command	*/
/*	extract the final status and come back	*/
/*	with the whole reply.			*/
/*						*/
/************************************************/
static int getreply(TLSFILE *tls,char **reply)

{
int status;
int proceed;

status=ERRPROC;
proceed=true;
while (proceed==true) {
  int sofar;
  char *ptr;
  char *buffer;

  if ((buffer=tls_waitforaction(MXTIME,tls))==(char *)0) {
    break;	/*time_out!		*/
    }
  ptr=buffer;
  proceed=false;
  if (sscanf(buffer,"%d%n",&status,&sofar)==1) {
    if (buffer[sofar]=='-')
      proceed=true;
    ptr=buffer+sofar+1;
    }
  (void) memmove(buffer,ptr,strlen(ptr)+1);
  (void) que_cleanstring(buffer);
  *reply=rou_memstrcat(*reply,"%s\n",buffer);
  buffer=rou_freestr(buffer);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to override remote IP number	*/
/*	with a new IP number.			*/ 
/*						*/
/************************************************/
static void doorgn(CTTYP *cta,char *nip)

{
typedef	struct hostent HTYP;

int t;
struct in_addr adr;
HTYP *h;
char *peername;
int phase;
int proceed;

t=sizeof(adr.s_addr);
h=(HTYP *)0;
peername=NORVERS;
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*error trap		*/
      proceed=false;
      break;
    case   1	:	/*checking good ipnum	*/
      if (strlen(nip)==0) {
        (void) eml_printfcrlf(cta->tls,eml_ascstat(BADPAR,"IP number missing"));
	phase=-1;
        }
      break;
    case   2	: 	/*getting host		*/
      if (inet_aton(nip,&adr)==0) {
        (void) eml_printfcrlf(cta->tls,eml_ascstat(BADPAR,"Not an IP number"));
	phase=-1;
        }
      break;
    case   3	:	/*getting host		*/
      if ((h=gethostbyaddr((const char *)&(adr.s_addr),t,AF_INET))!=(HTYP *)0){
	peername=h->h_name;
  	}
      break;
    case   4	:	/*setting contact value	*/
      (void) free(cta->peername);
      (void) free(cta->ipremote);
      cta->peername=strdup(peername);
      cta->ipremote=strdup(nip);
      break;
    case   5	:	/*updating process info	*/
      (void) eml_settitle(cta);
      (void) eml_printfcrlf(cta->tls,eml_ascstat(CMDOK));
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare the header add on	*/
/*	received information			*/
/*						*/
/************************************************/
static void mkreceived(FILE *file,CTTYP *contact)

{
time_t curtime;
int i;

curtime=rou_systime();
for (i=0;i<4;i++) {
  switch (i) {
    case 0	:
      (void) fprintf(file,"Received: from %s (%s [%s])\n",
			  contact->peername,contact->peername,
			  contact->ipremote);
      break;
    case 1	:
      (void) fprintf(file,"\tby %s (%s-%s) with SMTP\n",
                          contact->locname,appname,rou_getvers());
      break;
    case 2	:
      if (contact->cipherid!=(char *)0)
        (void) fprintf(file,"\t%s\n",contact->cipherid);
      break;
    case 3	:
      (void) fprintf(file,"\tid <%s@%s> %s",
                          contact->session,contact->locname,ctime(&curtime));
      break;
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an HELP explaination	*/
/*	to remote server.			*/
/*						*/
/************************************************/
static void dohelp(TLSFILE *tls,const VOCTYP *kwlst,char *cmd)


{
#define	HLPMRK	214

if (strlen(cmd)>0) {
  VOCTYP *voc;

  if ((voc=rou_getvoca(kwlst,cmd))!=(VOCTYP *)0) {
    ACTTYP *acting;

    acting=(ACTTYP *)(voc->ptr);
    if ((acting!=(ACTTYP *)0)&&(acting->help!=(char **)0)) {
      int i;

      for (i=0;acting->help[i]!=(char *)0;i++) {
        (void) eml_printfcrlf(tls,"%d- %s",HLPMRK,acting->help[i]);
        }
      }
    else {
      (void) eml_printfcrlf(tls,"%d HELP not available",HLPMRK);
      }
    (void) eml_printfcrlf(tls,"%d  End of HELP info",HLPMRK);
    } 
  else {
    (void) eml_printfcrlf(tls,"%d HELP topic \"%s\" unknown",HLPMRK,cmd);
    }
  }
else {
  int i;
  int cpt;
  char helpstr[200];

  (void) eml_printfcrlf(tls,"%d- This is %s version %s",HLPMRK,
			  appname,rou_getvers());
  (void) eml_printfcrlf(tls,"%d- Topics",HLPMRK);
  for (i=0,cpt=0;kwlst[i].key!=(char *)0;i++) {
    ACTTYP *acting;

    acting=(ACTTYP *)kwlst[i].ptr;
    if (acting->help!=(char **)0) {
      char dhelp[100];
      char help[100];

      if ((cpt%5)==0)
        (void) strcpy(helpstr,"");
      (void) snprintf(dhelp,sizeof(dhelp),"'%s'",kwlst[i].key);
      (void) snprintf(help,sizeof(help),"%-14s",dhelp);
      (void) strcat(helpstr,help);
      if ((cpt%5)==4) {
        (void) eml_printfcrlf(tls,"%d- %s",HLPMRK,helpstr);
        (void) strcpy(helpstr,"");
	}
      cpt++;
      }
    }
  if (strlen(helpstr)>0)
    (void) eml_printfcrlf(tls,"%d- %s",HLPMRK,helpstr);
  (void) eml_printfcrlf(tls,"%d- For more info use \"HELP <topic>\".",HLPMRK);
  (void) eml_printfcrlf(tls,"%d- To report bugs in the implementation "
			  "send email to",HLPMRK);
  (void) eml_printfcrlf(tls,"%d-\t%s@%s",HLPMRK,appname,SAFEDOM);
  (void) eml_printfcrlf(tls,"%d- For local information send email to "
			  "Postmaster at your site.",HLPMRK);
  (void) eml_printfcrlf(tls,"%d  End of HELP info",HLPMRK);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a domain name 	*/
/*	element is according the proper syntaxe	*/
/*						*/
/************************************************/
static int iselement(char *element)

{
int status;

status=false;
if (strlen(element)>0) {
  if (isalnum(element[0])!=0) {
    int i;

    status=true;
    for (i=1;i<strlen(element)-1;i++) {
      if (isalnum(element[i])!=0)
	continue;
      if (element[i]=='-')
	continue;
      status=false;
      break;
      }
    }
  if (element[0]=='#') {
    int i;

    status=true;
    for (i=1;i<strlen(element)-1;i++) {
      if (isdigit(element[i])!=0)
	continue;
      status=false;
      break;
      }
    }
  if (element[0]=='[') {
    if (element[strlen(element)-1]==']') {
      int i;

      status=true;
      for (i=1;i<strlen(element)-1;i++) {
	if (isdigit(element[i])!=0)
	  continue;
	if (element[i]=='.')
	  continue;
	status=false;
	break;
	}
      }
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an E-mail address	*/
/*	is correct.				*/
/*						*/
/************************************************/
static int isemailok(char *streml,char *errstr,int taille)

{
int status;

status=CMDOK;
if (strlen(streml)>0) {
  int proceed;
  int phase;
  char *localpart;
  char *domain;
  char *ptr;

  proceed=true;
  phase=1;
  localpart=strdup(streml);
  ptr=strchr(localpart,'@');
  domain=strdup(ptr+1);
  *ptr='\000';
  while (proceed==true) {
    switch (phase)  {
      case   0	:	/*error collector	*/
	status=BADPAR;
        proceed=false;
        break;
      case   1	:
	if (strlen(localpart)==0) {
	  (void) snprintf(errstr,taille,"Syntax error, missing localpart");
	  phase=-1;
	  }
	else {
	  int i;

	  for (i=0;(localpart[i]!='\000')&&(phase>0);i++) {
	    switch (localpart[i]) {
	      case ' '	:
	      case '<'	:
	      case '>'	:
	      case '('	:
	      case '['	:
	      case ']'	:
	      case ';'	:
	      case ':'	:
	        (void) snprintf(errstr,taille,"Syntax error, '%c' not allowed "
					      "in localpart",localpart[i]);
	        phase=-1;
		break;
	      default	:
		break;
	      }
	    }
	  }
        break;
      case   2	:
	if (strlen(domain)==0) {
	  (void) snprintf(errstr,taille,"Syntax error, missing domain part");
	  phase=-1;
	  }
	else {
	  char **elst;
	  char *dptr;
	  int i;

	  elst=(char **)0;
	  while ((dptr=strrchr(domain,'.'))!=(char *)0) {
	    if (dptr[strlen(dptr)-1]==']') {
	      dptr=strrchr(domain,'[');
	      if (dptr==(char *)0)
	        break;
	      if (dptr!=domain) {
		*(dptr-1)='\000';
		}
	      else
	        break;
	      }
	    else {
	      *dptr='\000';
	      dptr++;
	      }
	    elst=(char **)rou_addveclst((void **)elst,(void *)strdup(dptr));
	    }
	  elst=(char **)rou_addveclst((void **)elst,(void *)strdup(domain));
	  for (i=0;elst[i]!=(char *)0;i++) {
	     if (iselement(elst[i])!=true) {
	       char *bad;
	       int j,k;

	       bad=(char *)calloc(strlen(elst[i])*3,sizeof(char));
	       (void) strcpy(bad,elst[i]);
	       for (j=0,k=0;elst[i][j]!='\000';j++,k++) {
	         bad[k]=elst[i][j];
	         if (bad[k]=='%') {
	           k++;
		   bad[k]='%';
                   }
		 }
	       (void) snprintf(errstr,taille,"Syntax error, wrong domain "
					     "part '%s' ",bad);
	       (void) free(bad);
	       phase=-1;
	       break;
	       }
	     }
	  elst=rou_freestrlst(elst);
	  }
        break;
      default	:
        proceed=false;
        break;
      }
    phase++;
    }
  (void) free(domain);
  (void) free(localpart);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert "size=xxx" from	*/
/*	"MAIL FROM:" SMTP command, return	*/
/*	the mail size value.			*/
/*						*/
/************************************************/
int gettaille(char *buffer)

{
u_long mtaille;
char *ptr;

mtaille=0;
if ((ptr=strstr(buffer,"size"))!=(char *)0) {
  if ((ptr=strchr(ptr,'='))!=(char *)0) {
    ptr++;
    (void) sscanf(ptr,"%lu",&mtaille);
    }
  }
return mtaille;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract and verify email	*/
/*	addresse from the SMTP protocol exchange*/
/*	Returne the E-mail address within a	*/
/*	dynamically allocated memory, buffer	*/
/*	variable is updated with the remaining	*/
/*	data.					*/
/*						*/
/************************************************/
static char *extract(char *buffer,char *errstr,int taille)

{
char *email;
int proceed;
int phase;

email=(char *)0;
proceed=true;
phase=1;
while (proceed==true) {
  char *ptr;

  ptr=(char *)0;
  switch (phase) {
    case   0	:	/*error collector	*/
      if (email!=(char *)0) {
	(void) free(email);
	email=(char *)0;
        }
      proceed=false;
      break;
    case   1	:
      if ((ptr=strchr(buffer,'<'))!=(char *)0) {
	ptr++;
	(void) memmove(buffer,ptr,strlen(ptr)+1);
	}
      else {
	(void) snprintf(errstr,taille,"Syntax error, '<' not found");
	phase=-1;
	}
      break;
    case   2	:
      if ((ptr=strchr(buffer,'>'))!=(char *)0) {
	*ptr='\000';
	email=strdup(buffer);
	ptr++;
	while (*ptr==' ')
	  ptr++;
        (void) memmove(buffer,ptr,strlen(ptr)+1);/*getting remaining	*/
        }
      else {
	(void) snprintf(errstr,taille,"Syntax error, '>' not found");
	phase=-1;
	}
      break;
    case   3	:
      if ((ptr=strchr(email,':'))!=(char *)0) {
	ptr++;
	(void) memmove(email,ptr,strlen(ptr)+1);
	}
      break;
    case   4	:
      if (strlen(email)==0)
	phase=999;	/*empty email ok if it is a FROM:	*/
      break;
    case   5	:	/*standardizing email case		*/
      (void) rou_strtolower(email);
      break;
    case   6	:
      if ((ptr=strchr(email,'@'))!=(char *)0) {
	if (strrchr(email,'@')!=ptr) {
	  (void) snprintf(errstr,taille,
			  "Syntax error, too many '@' charactere");
	  phase=-1;
          }
	}
      else {
	(void) snprintf(errstr,taille,
			"Syntax error, missing '@' charactere");
	phase=-1;
	}
      break;
    case   7	:	/*checking if there wrong char in email	*/
      if (email!=(char *)0) {
	char *ptre;

        ptre=email;
	while ((*ptre!='\000')&&(phase>0)) {
	  switch (*ptre) {
	    case ' '	:
	      (void) snprintf(errstr,taille,
			      "Syntax error, wrong char within Email");
	      phase=-1;
	      break;
	    default	:
	      break;
	    }
	  ptre++;
	  }
        }
      break;
#ifdef BUGRFC
    case   8	:	/*removing "'" type caractere		*/
      if (strchr(email,'\'')!=(char *)0)
	phase=-1;
      break;
#endif
    default	:	/*Everything fine			*/
      proceed=false;
      break;
    }
  phase++;
  }
return email;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check an ID provided to 	*/
/*	further access the clement (and mentor)	*/
/*	WEB interface.				*/
/*						*/
/************************************************/
static char *checkid(CTTYP *contact,char *buffer)

{
char *id;
int phase;
int proceed;

id=(char *)0;
phase=1;
proceed=true;
while (proceed==true) {
  char *ptr;

  ptr=(char *)0;
  switch (phase) {
    case    0	:	/*phase out			*/
      id=rou_freestr(id);
      proceed=false;
      break;
    case    1	:	/*check syntaxe			*/
      if ((ptr=strchr(buffer,':'))!=(char *)0) {
        (void) memmove(buffer,ptr,strlen(ptr)+1);
	}
      else {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(SYNERR,"':' expected"));
	phase=-1;
        }
      break;
    case    2	:	/*extracting ID information	*/
      id=strdup(buffer);
      break;
    default	:	/*Safe Guard, everything fine	*/
      proceed=false;
      break;
    }
  phase++;
  }
return id;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to count the number of	*/
/*	recipient in queue and entitled to 	*/
/*	receive data E-mail.			*/
/*						*/
/************************************************/
static int countrcpts(RCPTYP **rcpts)

{
int count;

count=0;
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    switch (rcpts[i]->status) {
      case m_indata	:
      case m_relayed	:
	count++;
	break;
      default		:
        break;
       }
    }
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an "EHLO" message	*/
/*	an collect working parameters		*/
/*						*/
/************************************************/
static int doehlo(CTTYP *contact)

{
#define	WEHLO	20	/*20 sec for EHLO reply	*/

typedef	enum	{
	h_tls,		/*STARTTLS capable	*/
	h_rqstid,	/*RQSTID capable	*/
	h_unknown	/*unknown code		*/
	}HTYP;

static VOCTYP vochelo[]={
        {h_tls,STARTTLS,(const void *)0},
        {h_rqstid,RQSTID,(const void *)0},
        {h_unknown,(char *)0,(const void *)0}
        };

int done;
int proceed;

done=true;
proceed=true;
(void) eml_printfcrlf(contact->tls,EHLO,contact->locname);
while (proceed==true) {
  int numret;
  char *ptr;
  VOCTYP *voc;
  int sofar;
  char *buffer;

  numret=0;
  ptr=(char *)0;
  voc=(VOCTYP *)0;
  if ((buffer=tls_waitforaction(MXTIME,contact->tls))==(char *)0) {
    done=false;
    break;
    }
  if (sscanf(buffer,"%d%n",&numret,&sofar)!=1) {
    done=false;
    break;
    }
  if (buffer[sofar]==' ') /*end of EHLO	*/
    proceed=false;
  ptr=buffer+sofar+1;
  (void) memmove(buffer,ptr,strlen(ptr)+1);
  (void) que_cleanstring(buffer);
  if ((voc=rou_getvoca(vochelo,buffer))!=(VOCTYP *)0) {
    switch (voc->code) {
      case h_tls	:
	contact->capable|=CPTLS;
	break;
      case h_rqstid	:
	contact->capable|=CPRQT;
	break;
      default		:
        break;
      }
    }
  buffer=rou_freestr(buffer);
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait response from remote	*/
/*	checker.				*/
/*						*/
/************************************************/
static char *gettlsline(char *line,int taille,CTTYP *contact)

{
char *ptr;

ptr=(char *)0;
if (tls_fgets(line,taille,contact->tls)==(char *)0) {
  switch(tls_waitforchar(contact->tls,1000)) {
    case -1		:
      switch(errno) {
	case EINTR	:	/*received a signal, lets see...*/
	  break;		/*must be an IPC messages	*/
	default		:	/*hummm real code fault report	*/
	  (void) code_fault("%s select error '%s'",
			    "modchk.c:gettlsline",strerror(errno));
	  break;
	}
      break;
    case  0		:
     				/*standard time out		*/
      break;
    default		:
      if (tls_getfeed(contact->tls)==false)
        quickout=true;		/*trouble or EOF		*/
      if (tls_iseof(contact->tls)==true)
        quickout=true;
      break;
    }
  }
else {
  (void) que_cleanstring(line);
  ptr=line;
  }
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send the DROPME code to the*/
/*	advising Mentor.			*/
/*						*/
/************************************************/
static void senddropme(CHKTYP *checker)

{
if ((checker!=(CHKTYP *)0)&&(checker->contact!=(CTTYP *)0)) {
  (void) eml_printfcrlf(checker->contact->tls,"%s %s\n%s\n.",
					       RQSTID,ONEWAY,DROPME);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send the CKCODE		*/
/*	authentication to renmote Checker, send	*/
/*	The current listening IP too.		*/
/*						*/
/************************************************/
static int sendckcode(SKRTYP *skr,char *uniqid,CTTYP *cta)

{
#define	WTCKCODE	20	/*20 seconds waiting time	*/

int succeed;
char **list;
char *msgref;
char **resp;
time_t curtime;
int phase;
int proceed;

succeed=false;
list=skr->bindlst;
msgref=(char *)0;
resp=(char **)0;
curtime=(time_t)0;
phase=1;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case  0	:		/*phase error		*/
      proceed=false;
      break;
    case  1	:		/*sending CKCODE value	*/
      (void) eml_printfcrlf(cta->tls,"%s %s\n%s %s:%s = %d",
				     RQSTID,uniqid,CKCODE,skr->chkcode,VMARK,rou_getnumvers());
      break;
    case  2	:		/*sending binding info	*/
      if (list!=(char **)0) {
        int i;

        for (i=0;list[i]!=(char *)0;i++) {
	  char *binder;
	  char *line;

	  binder=strdup(list[i]);	
	  line=binder;
	  while (line!=(char *)0) {
	    char *ptr;
	    char *comma;

	    if ((ptr=strchr(line,':'))!=(char *)0) {
	      *ptr='\000';
	      ptr++;
	      }
            if ((comma=strchr(line,','))!=(char *)0) {
	      char *seccom;

	      if ((seccom=strchr(comma+1,','))!=(char *)0)
	        *seccom='\000';
	      }	
            (void) eml_printfcrlf(cta->tls,"%s",line);
	    line=ptr;
	    }
	  binder=rou_freestr(binder);
	  }
	}
      break;
    case  3	:	/*closing the CKCODE request	*/
      (void) eml_printfcrlf(cta->tls,".");
      break;
    case  4	:	/*waiting for answer 20 sec max	*/
      curtime=rou_systime();
      while ((curtime+WTCKCODE)>rou_systime()) {
	char line[100];

	if (gettlsline(line,sizeof(line),cta)!=(char *)0) {
	  resp=(char **)rou_addveclst((void **)resp,(void *)strdup(line));
	  if ((strlen(line)==1)&&(line[0]=='.'))
	    break;
	  }
	}
      break;
    case  5	:	/*checking for time out		*/
      if ((resp==(char **)0)||(rou_vecsize((void **)resp)!=3)) {
        (void) journalling(false,APJRL,"CKCODE noanswer from checker");
	}
      else {		/*checking for answer		*/
	int scancod;	

	if (sscanf(resp[1],"%d",&scancod)==1) {
	  if (scancod==CMDOK)	
	    succeed=true;
	  else
            (void) journalling(false,APJRL,"Checker refuse authentication "
					   "(error=<%s>",resp[1]);
	  }
	resp=rou_freestrlst(resp);
	}
    default	:	/*SAFE guard phase		*/
      proceed=false;
      break;
    }
  phase++;
  }
return succeed;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assign CHKTYP memory	*/
/*	structure. Open the channel with the	*/
/*	Remote SMTP server.			*/
/*						*/
/************************************************/
static CHKTYP *openremotechecker(SKRTYP *skr,char *uniqid)

{
#define	OPCK	"lvleml.c:openremotechecker,"
#define	CKWAIT	20	/*MAX Waiting time	*/


CHKTYP *checker;
char *scan;
int phase;
int proceed;

checker=(CHKTYP *)0;
scan=strdup(skr->hostname);
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*properly exit if trouble	*/
      checker=eml_closesmtp(checker);
      proceed=false;
      break;
    case   1	:
      if ((checker=eml_opensmtp(scan,CKWAIT))==(CHKTYP *)0) {
        (void) journalling(false,APJRL,"Unable to contact checker <%s>",scan);
        (void) alert(0,"%s Unable to access checking server <%s>",OPCK,scan);
	phase=9999;	/*trouble trouble get out fast	*/
	}
      break;
    case   2	:	/*Is remote a CHECKER		*/
      if (((checker->contact->capable&CPRQT)==0)||(skr->chkcode==(char *)0))
	phase++;	/*do NOT send CKCODE		*/
      break;
    case   3	:	/*Sending CKCODE to checker	*/
      if (sendckcode(skr,uniqid,checker->contact)!=true) {
        (void) journalling(false,APJRL,"Checker <%s> No authentication",scan);
        (void) alert(0,"%s Checker <%s> refuse our authentication",OPCK,scan);
	phase=-1;	/*not a good status		*/
	}
      break;
    case   4	:	/*everything fine, journalling	*/
      if (debug>0)
        (void) journalling(false,APJRL,"Checker <%s> connected",scan);
      break;
    default	:	/*safe guard everything fine	*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(scan);
return checker;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send the data-contents	*/
/*	to local sendmail daemon.		*/
/*						*/
/************************************************/
static int sendsmtp(CTTYP *contact,char *email,char *emailname,char **info)

{
#define	OPOX	"lvleml.c:sendsmtp,"

int done;
int recipient;
int proceed;
int i;

done=false;
recipient=-1;
proceed=true;
for (i=0;info[i]!=(char *)0;i++) {
  int retstat;
  char *str;

  str="RCPT TO: <%s>";
  switch (i) {
    case 0	:
      str="MAIL FROM: <%s>";
      if (strcasecmp(MAILDEM,info[i])==0)
	(void) strcpy(info[i],"");
      break;
    default	:
      break;
    }
  recipient++;
  if ((retstat=eml_sendcmd(contact,CMDOK,str,info[i]))!=CMDOK) {
    if (i==0) 
      break;		/*this should never ever occure	*/
    recipient--;	/*do not count the recipient	*/
    (void) journalling(false,EVJRL,"\t %s,  recipient <%s> not accepted",
				    email,info[i]);
    }
  }
if (recipient<=0) {
  proceed=false;
  done=false;
  }
if (proceed==true) {
  FILE *fdata;

  if ((fdata=fopen(emailname,"r"))!=(FILE *)0) {
    if (eml_sendcmd(contact,DATAOK,DATA)==DATAOK) {
      char datain[2000];

      /*ATTENTION transmission avec CRLF OBLIGATOIRE	*/
      while (fgets(datain,sizeof(datain)-1,fdata)!=(char *)0) {
        register int taille;

	if (datain[0]=='.') {
	  if ((datain[1]=='\r')||(datain[1]=='\n'))
	  (void) strcpy(datain,"..\n");
	  }
	taille=strlen(datain);
	if (taille>0) {
	  taille--;
	  if ((datain[taille]=='\n')&&(datain[taille-1]!='\r')) {
	    datain[taille]='\r';
	    datain[taille+1]='\n';
	    datain[taille+2]='\000';
	    }
	  }
        (void) tls_write(contact->tls,datain,strlen(datain));
	}
      if (eml_sendcmd(contact,CMDOK,".")!=CMDOK) {
        (void) alert(0,"%s data contents not accepted email-id=<%s> "
		       "got status <%s>",OPOX,email,contact->info);
	}
      else
	done=true;	/*file was transmitted	*/
      }
    (void) fclose(fdata);
    }
  }
else 
  (void) alert(0,"%s Command contents not accepted for email-id=<%s> "
		 "got status <%s>",OPOX,email,contact->info);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dispatch a file (E-mail) to*/
/*	an auxiliary server or to store it	*/
/*	within the local E-mail spooler (if 	*/
/*	configured as "keeplocal")		*/
/*						*/
/************************************************/
static int dispatching(CHKTYP *auxi,char *queudir,char *emailid)

{
#define	OPLD	"lvleml.c:dispatching,"

int done;
MLTYP *ml;

done=false;
if ((ml=que_getemails(queudir,emailid))!=(MLTYP *)0) {
  char **infos;
  char *mode;
  char emailname[300];
  char emailqueue[300];

  done=true;
  infos=(char **)0;
  mode="Relaying";
  if (auxi->control==c_local)
    mode="Storing";
  (void) journalling(false,EVJRL,"%s %s -> %s",mode,emailid,auxi->hostname);
  (void) rou_mkname(emailname,sizeof(emailname),d_spool,
						queudir,emailid,EMAILEXT);
  (void) rou_mkname(emailqueue,sizeof(emailqueue),d_spool,
						  queudir,emailid,QUEUEEXT);
  infos=(char **)rou_addveclst((void **)infos,(void *)ml->from);
  if (ml->rcpts!=(char **)0) {
    int i;

    for (i=0;ml->rcpts[i]!=(char *)0;i++) {
      if ((ml->spamperti!=(char *)0)&&(ml->rcpts[i]!=(char *)0)) {
	char *webgood;
        char *domain;
	DFSTYP *dfs;

        webgood=(char *)0;
        domain=eml_strdomain(ml->rcpts[i]);
        if ((dfs=eml_locatedfs(auxi->dfs,domain))!=(DFSTYP *)0) {
          webgood=dfs->webgood;
	  }
        (void) spammark(ml,webgood,queudir,emailid,
			ml->spamperti,ml->rcpts[i]);
        }
      infos=(char **)rou_addveclst((void **)infos,(void *)ml->rcpts[i]);
      }
    }
  if (debug>0)
    (void) fprintf(stdout,"##-%s %s <%s>\n",
		          rou_getstrtime(rou_systime()),mode,emailid);
  switch (auxi->control) {
    case c_standard	:
      done=sendsmtp(auxi->contact,emailid,emailname,infos);
      break;
    case c_local	:
      done=que_storeemails(infos[0],infos+1,emailname,emailid);
      break;
    case c_mentor	:
      (void) alert(0,"%s Unexpected control (bug?)",OPLD);
      break;
    }
  (void) free(infos);	/*info pointer contents were NOT duplicated	*/
  if (done==false) { 	/*storing problem in the trouble queue		*/
    char qemailname[300];
    char qemailqueue[300];
    char trbldir[250];

    (void) snprintf(trbldir,sizeof(trbldir),"%s%s",TRBLDIR,queudir);
    (void) rou_mkname(qemailname,sizeof(qemailname),d_spool,trbldir,"","");
    (void) rou_createdirs(qemailname);
    (void) journalling(false,EVJRL,"%s %s -> %s","Trouble",emailid,"trblq");
    (void) alert(0,"Moving email-id=<%s> to trouble dir queue <%s>",
		   emailid,qemailname);
    (void) rou_mkname(qemailname,sizeof(qemailname),d_spool,
						    trbldir,emailid,EMAILEXT);
    (void) rou_mkname(qemailqueue,sizeof(qemailqueue),d_spool,
						    trbldir,emailid,QUEUEEXT);
    if (link(emailname,qemailname)<0)
      (void) alert(0,"%s Unable to link file name <%s> to <%s> (bug?)",
		      OPLD,emailname,qemailname);
    if (link(emailqueue,qemailqueue)<0)
      (void) alert(0,"%s Unable to link file name <%s> to <%s> (bug!?)",
		      OPLD,emailqueue,qemailqueue);
    done=true;	/*lets transmit the next one	*/
    }
  (void) unlink(emailqueue);
  (void) unlink(emailname);
  ml=que_dropemails(ml);
  (void) journalling(true,EVJRL,"");
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract authentication	*/
/*	info.					*/
/*						*/
/************************************************/
static char *getauth(TLSFILE *tls,char *buffer)

{
#define MXWLOG	60

static char *vocloc[]={"plain","login",(char *)0};
static char *logdat[]={"VXNlcm5hbWU6", /*Username: in B64	*/
		       "UGFzc3dvcmQ6", /*Password: in B64	*/
		       (char *)0};

char *info;
char **login;
char *ptr;
int code;
int i;
char auth[200];

info=(char *)0;
login=(char **)0;
(void) strncpy(auth,buffer,sizeof(auth)-2);
if ((ptr=strchr(buffer,' '))!=(char *)0) {
  *ptr='\000';
  (void) strncpy(auth,buffer,sizeof(auth)-2);
  ptr++;
  }
else
  ptr="";
(void) memmove(buffer,ptr,strlen(ptr)+1);
for (code=0;vocloc[code]!=(char *)0;code++) {
  if (strcasecmp(auth,vocloc[code])==0) {
    break;
    }
  }
switch(code) {
    case  0  	:
      login=tls_getb64(login,buffer);
      break;
    case  1  	:
      i=0;
      login=(char **)rou_addveclst((void **)login,(void *)strdup(""));
      if (strlen(buffer)>0) {
        login=tls_getb64(login,buffer);
	i=1;
	}
      for (;i<2;i++) {
        char *strloc;

        (void) eml_printfcrlf(tls,"%d %s",SENDB64,logdat[i]);
        if ((strloc=tls_waitforaction(MXTIME,tls))==(char *)0) {
	  login=rou_freestrlst(login);
	  break;
	  }
	else {
          login=tls_getb64(login,strloc);
	  }
        rou_freestr(strloc);
	}
      break;
    default  	:
      (void) alert(0,"mngeml.c:checkauth, Unknown <%s> SMTP "
		     "authentication mode",auth);
      (void) eml_printfcrlf(tls,"%s",eml_ascstat(BADAUTH,auth));
      break;
    }
if (login!=(char **)0) {
  int taille;

  for (taille=0,i=0;login[i]!=(char *)0;i++)
     taille+=strlen(login[i]);
  taille+=20;
  info=calloc(taille,sizeof(char));
  (void) snprintf(info,taille,"<%s> %s void",login[1],login[2]);
  login=rou_freestrlst(login);
  }
return info;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check user authentication	*/
/*	IF authentication capability is enabled	*/
/*						*/
/************************************************/
static int checkviaauth(int TLSauth,CTTYP *cta,char *buffer)

{
int trouble;
int phase;
int proceed;

trouble=true;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*already authenticated?*/
      if (cta->authok==true) {
	const char *ptr;

	ptr=eml_ascstat(BADSEQ,"Already Authenticated");
        (void) usleep(100000);
        (void) eml_printfcrlf(cta->tls,"%s",ptr);
	phase=999;	/*no need to go further	*/
        }
      break;
    case 1	:	/*already authenticated?*/
      if ((TLSauth==true)&&(cta->cipherid==(char *)0)) {
	const char *ptr;

	ptr=eml_ascstat(BADSEQ,"Authentication request not permited");
        (void) usleep(100000);
        (void) eml_printfcrlf(cta->tls,"%s",ptr);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 2	:	/*check authentication	*/
      cta->info=rou_freestr(cta->info);
      if ((cta->info=getauth(cta->tls,buffer))==(char *)0)
	phase=999;	/*no need to go further	*/
      break;
    case 3	:	/*everything fine	*/
      trouble=false;
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return trouble;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a remote serveur	*/
/*	now about a specific user.		*/
/*						*/
/************************************************/
static int checkremoteuser(CHKTYP *checker,char *email)

{
int status;
char *domain;
int phase;
int proceed;

status=ERRPROC;
domain=eml_strdomain(email);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*in Sync with remote	*/
      if (eml_sendcmd(checker->contact,CMDOK,RSET)!=CMDOK) 
        proceed=false;
      break;
    case 1	:	/*sending "Mail FROM"	*/
      if (eml_sendcmd(checker->contact,CMDOK,MAILFRM,"")!=CMDOK) 
        proceed=false;
      break;
    case 2	:	/*sending "RCPT TO"	*/
      status=eml_sendcmd(checker->contact,CMDOK,RCPTTO,email);
      break;
    case 3	:	/*are we using the relay*/
      if ((checker->relay==true)&&(status=CMDOK))
	status=RELAYOK;	/*allowing many recipient	*/
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate an already open	*/
/*	remote within the remote list		*/
/*	to the IPC channel.			*/
/*						*/
/************************************************/
static SKRTYP *locremote(SKRTYP **skrlst,char *domain,CKPREF pref)

{
SKRTYP *skr;

skr=(SKRTYP *)0;
if (skrlst!=(SKRTYP **)0) {
  int i;

  for (i=0;skrlst[i]!=(SKRTYP *)0;i++) {
    if (skrlst[i]->pref!=pref)
      continue;
    if (eml_locatedfs(skrlst[i]->dfs,domain)!=(DFSTYP *)0) {
      skr=skrlst[i];
      break;
      }
    }
  }
return skr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to establish a TSL as remote	*/
/*	is TLS capable.				*/
/*						*/
/************************************************/
static void establishTLS(CTTYP *cta,char *sendmail)

{
#define	OPES	"lvleml.c:establishTLS,"

CNFTYP *cnf;
int phase;
int proceed;

cnf=(CNFTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0 	:	/*sending STARTTLS rqst	*/
      if (eml_sendcmd(cta,SIGNON,STARTTLS)!=SIGNON)
	phase=999;	/*request not accepted!	*/
      break;
    case 1	:	/*reloading Config	*/
      if ((cnf=cnf_readconfig())==(CNFTYP *)0) {
        (void) alert(0,"%s Unable to load configuration!",OPES);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*initiating TLS	*/
      (void) tls_gotls(cta->tls,false,cnf->SSLciphrs);
      cnf=cnf_freeconfig(cnf);
      break;
    case 3	:	/*doing helo again	*/
      if (doehlo(cta)==false) {	/*Going EHLO Again	*/
        (void) alert(0,"%s (pid=%05d) Unable to carry TLS EHLO with '%s'",
		        OPES,getpid(),sendmail);
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a user is really 	*/
/*	existing, if the domain is remote,	*/
/*	a remote server is used via the SMTP	*/
/*	protocole.				*/
/*						*/
/************************************************/
int eml_checkuser(CHKTYP *checker,char *caller,char *rqst,
		  char *email,int relais,PLDTYP *pld)

{
#define OPCH	"lvleml.c:eml_checkuser,"

char **ptr;

pld->info.smstat.status=ERRPROC;
ptr=&(pld->info.smstat.str);
*ptr=(char *)0;
switch (checker->control) {
  case (c_local)	:
    pld->info.smstat.status=que_chkemails(email);
    switch(pld->info.smstat.status) {
      case NORELAY	:
	*ptr=strdup("Local domain not existing");
	break;
      case REJECT	:
	*ptr=strdup("Local mailbox not existing");
	break;
      case CMDOK	:
	*ptr=strdup("Requested mail action okay");
	if (relais==true)
          pld->info.smstat.status=RELAYOK;
	break;
      case ERRPROC	:
      default		:
	*ptr=rou_memstrcat((char *)0,"%s Unexpected status '%d'",
				      OPCH,pld->info.smstat.status);
	break;
      }
    break;
  case (c_standard)	:
    pld->info.smstat.status=checkremoteuser(checker,email);
    if (checker->contact->info!=(char *)0)
      *ptr=strdup(checker->contact->info);
    break;
  case (c_mentor)	:
    (void) alert(0,"lvleml.c:eml_checkuser, can't check with mentor (bug!)");
    break;
  }
return pld->info.smstat.status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "MAIL FROM:" 	*/
/*	command from the SMTP client.		*/
/*						*/
/************************************************/
int eml_checkfrom(CTTYP *contact,char *buffer)

{
int fromok;
int phase;
int proceed;
char *from;
char *ptr;
int mtaille;
char errstr[200];

fromok=false;
from=(char *)0;
ptr=(char *)0;
mtaille=0;
(void) strcpy(errstr,"");
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case    0	:	/*phase out			*/
      proceed=false;
      break;
    case    1	:	/*previous from?		*/
      if (contact->frm!=(FRMTYP *)0) {
        (void) eml_printfcrlf(contact->tls,
			      eml_ascstat(BADPAR,"Sender already specified"));
	phase=-1;
	}
      break;
    case    2	:	/*checking for ':'		*/
      if (strchr(buffer,':')==(char *)0) {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(SYNERR,"':' expected"));
	phase=-1;
        }
      break;
    case    3	:	/*checking From format		*/
      if ((from=extract(buffer,errstr,sizeof(errstr)))==(char *)0) {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(BADPAR,errstr));
        phase=-1;
        }
      break;
    case    4	:	/*checking from format		*/
      if (isemailok(from,errstr,sizeof(errstr))!=CMDOK) {
	(void) journalling(false,EVJRL,"wrong domain part <%s>",from);
        (void) eml_printfcrlf(contact->tls,eml_ascstat(BADPAR,errstr));
        phase=-1;
        }
      break;
    case    5	:	/*checking if originator empty	*/
      if (strlen(from)==0) {
	if (contact->heloname!=(char *)0) {
	  from=rou_freestr(from);
          from=rou_memstrcat(from,"%s@%s",MAILDEM,contact->heloname);
	  }
        else {
	  (void) snprintf(errstr,sizeof(errstr),"helo/ehlo missing hostname");
	  (void) journalling(false,EVJRL,"No domain part and empty 'from'");
          (void) eml_printfcrlf(contact->tls,eml_ascstat(BADPAR,errstr));
          phase=-1;
	  }
        }
      break;
    case    6	:	/*extracting mail size		*/
      mtaille=gettaille(buffer);
      break;
    default	:	/*Safe Guard, everything fine	*/
      contact->frm=(FRMTYP *)calloc(1,sizeof(FRMTYP));
      contact->frm->from=from;
      contact->frm->ipremote=strdup(contact->ipremote);
      contact->frm->listenip=strdup(contact->listenip);
      contact->frm->listenport=contact->listenport;
      contact->frm->mtaille=mtaille;
      from=(char *)0;
      fromok=true;
      proceed=false;
      break;
    }
  phase++;
  }
if (from!=(char *)0)
  (void) free(from);
return fromok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "RCPT TO:"	*/
/*	command from the SMTP client.		*/
/*						*/
/************************************************/
char *eml_checkto(CTTYP *contact,char *buffer)

{
char *to;
char *locbuf;
int phase;
int proceed;
char errstr[200];

to=(char *)0;
locbuf=strdup(buffer);
(void) strcpy(errstr,"");
phase=1;
proceed=true;
while (proceed==true) {
  int status;

  status=BADPAR;
  switch (phase) {
    case    0	:	/*phase out			*/
      to=rou_freestr(to);
      proceed=false;
      break;
    case    1	:	/*previous from?		*/
      if (contact->frm==(FRMTYP *)0) {
        (void) eml_printfcrlf(contact->tls,
			      eml_ascstat(status,"Sender must be specified"));
	phase=-1;
	}
      break;
    case    2	:	/*checking for ':'		*/
      if (strchr(locbuf,':')==(char *)0) {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(SYNERR,"':' expected"));
	phase=-1;
        }
      break;
    case    3	:	/*testing against <POSTMASTER>	*/
      if (strstr(locbuf,"<"POSTMASTER">")!=(char *)0) {
	char *locdom;
	
	if ((locdom=strchr(contact->locname,'.'))==(char *)0) {
          (void) journalling(true,APJRL,"Warning IP [%s] without good good reverser addresss <%s>",
					contact->listenip,contact->locname);
	  }
	else {
	  char *newloc;
	  register int dummycnt;

          (void) journalling(false,EVJRL,"RCPT TO: <%s> ->> adding domain <@%s> to it",
					POSTMASTER,locdom+1);
	  dummycnt=asprintf(&newloc,": <%s@%s>",POSTMASTER,locdom+1);
	  (void) free(locbuf);
	  locbuf=newloc;
	  }
	}
      break;
    case    4	:	/*extracting address format	*/
      if ((to=extract(locbuf,errstr,sizeof(errstr)))==(char *)0) {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(status,errstr));
        phase=-1;
        }
      break;
    case    5	:	/*could the address an empty one	*/
      if (strlen(to)==0) {
        const char *ptr;

	ptr=eml_ascstat(status,"Recipient must be specified");
        (void) eml_printfcrlf(contact->tls,ptr);
	phase=-1;
	}
      break;
    case    6	:	/*checking adresse format		*/
      if ((status=isemailok(to,errstr,sizeof(errstr)))!=CMDOK) {
        (void) eml_printfcrlf(contact->tls,eml_ascstat(status,errstr));
        phase=-1;
        }
      break;
    default	:	/*Safe Guard, everything fine	*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(locbuf);
return to;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "HELO" commande	*/
/*	from the SMTP client.			*/
/*						*/
/************************************************/
int eml_checkhelo(CTTYP *c,char *buffer,char sep)

{
int proceed;

proceed=true;
if (strlen(buffer)==0) {
  const char *ptr;

  proceed=false;
  ptr=eml_ascstat(BADPAR,"HELO requires domain address");
  (void) eml_printfcrlf(c->tls,ptr);
  }
else {
  c->heloname=rou_freestr(c->heloname);
  c->heloname=strdup(buffer);
  (void) eml_printfcrlf(c->tls,"%d%c%s Hello %s [%s], please to meet you",
			    CMDOK,sep,c->locname,c->peername,c->ipremote);
  }
return proceed;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display connection status	*/
/*	if doing command 'ps ax' from console	*/
/*						*/
/************************************************/
void eml_settitle(CTTYP *contact)

{
(void) settitle("%s reading on [%s:%d] since %s, from '%s/[%s]'",
                appname,contact->listenip,contact->listenport,contact->attime,
		contact->peername,contact->ipremote);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to count the number of user	*/
/*	receiving a direct E-mail.		*/
/*						*/
/************************************************/
int eml_countadvice(CTTYP *contact,MENTYP status)

{
int count;

count=0;
if ((contact!=(CTTYP *)0)&&(contact->rcpts!=(RCPTYP **)0)) {
  int i;

  for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
    if (contact->rcpts[i]->status==status)
      count++;
    }
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reset contact structure	*/
/*	to be ready for a new E-mail.		*/
/*						*/
/************************************************/
void eml_resetcontact(CTTYP *contact)

{
(void) unlink(contact->dataname);
contact->evl=hdr_freeevl(contact->evl);
contact->headers=rou_freestrlst(contact->headers);
contact->dataname=rou_freestr(contact->dataname);
contact->rcpts=cleanrcpts(contact->rcpts);
contact->frm=cleanfrm(contact->frm);
contact->lastxchg=rou_systime();
contact->dfltcodeok=false;
contact->sequence%=99;
contact->sequence++;
(void) journalling(true,EVJRL,(char *)0);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to initiate a contact array	*/
/*	With a remote client under the SMTP	*/
/*	proctol.				*/
/*						*/
/************************************************/
CTTYP *eml_gotcontact(TLSFILE *tls)

{
CTTYP *contact;
char dbgname[400];


contact=(CTTYP *)calloc(1,sizeof(CTTYP));
contact->tls=tls;
contact->lastxchg=rou_systime();
contact->sequence=1;
contact->session=rou_uniquename();
contact->ckcode=strdup(NCKCODE);
contact->info=(char *)calloc(10,sizeof(char));
contact->locname=tls_getaddrname(tls,pr_locname,"NO_HOST_NAME");
contact->listenip=tls_getaddrname(tls,pr_locip,"?.?.?.?");
contact->peername=tls_getaddrname(tls,pr_remname,NORVERS);
contact->ipremote=tls_getaddrname(tls,pr_remip,"?.?.?.?");
(void) rou_mkname(dbgname,sizeof(dbgname),d_spool,DBGSDIR,
					  contact->session,".dbgs");
contact->dbgname=strdup(dbgname);
contact->prvdbg=debug;
if ((stdout=freopen(contact->dbgname,"w",stdout))==(FILE *)0)
  (void) core_dump("mngeml.c:eml_gotcontact, Unable to reopen stdout "
		   "(error=<%s>)",strerror(errno));
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open and SMTP connection	*/
/*	according server information		*/
/*						*/
/************************************************/
CHKTYP *eml_opensmtp(char *server,int ckwait)

{
#define	OPSR	"lvleml.c:eml_opensmtp,"

CHKTYP *check;

check=(CHKTYP *)0;
if (server!=(char *)0) {
  int port;
  CTTYP *contact;
  char hostname[200];

  contact=(CTTYP *)0;
  if (sscanf(server,"%95[^\n\r,],%d",hostname,&port)!=2) {
    (void) alert(0,"%s Unable to scan <%s>",OPSR,server);
    }
  else {
    if ((contact=eml_openmails(hostname,port,ckwait))!=(CTTYP *)0) {
      (void) tracing(TRALOG,10,"Got contact with checker <%s>",hostname);
      check=(CHKTYP *)calloc(1,sizeof(CHKTYP));
      check->ckwait=ckwait;
      check->hostname=strdup(hostname);
      check->contact=contact;
      }
    }
  }
return check;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to terminate an SMTP connexion*/
/*						*/
/************************************************/
CHKTYP *eml_closesmtp(CHKTYP *smtp)

{
if (smtp!=(CHKTYP *)0) {
  if (smtp->contact!=(CTTYP *)0) {
    (void) eml_sendcmd(smtp->contact,QUITOK,"QUIT");
    (void) usleep(10000);
    (void) tls_close(smtp->contact->tls);
    smtp->contact=eml_itsover(smtp->contact);
    }
  smtp->hostname=rou_freestr(smtp->hostname);
  smtp->dfs=eml_freedfslst(smtp->dfs);
  (void) free(smtp);
  smtp=(CHKTYP *)0;
  }
return smtp;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory use by an 	*/
/*	SMTP contact.				*/
/*						*/
/************************************************/
CTTYP *eml_itsover(CTTYP *contact)

{
if (contact!=(CTTYP *)0) {
  (void) eml_resetcontact(contact);
  contact->rcpts=cleanrcpts(contact->rcpts);
  contact->frm=cleanfrm(contact->frm);
  contact->ckcode=rou_freestr(contact->ckcode);
  contact->clamvers=rou_freestr(contact->clamvers);
  contact->cipherid=rou_freestr(contact->cipherid);
  contact->ipremote=rou_freestr(contact->ipremote);
  contact->peername=rou_freestr(contact->peername);
  contact->hostname=rou_freestr(contact->hostname);
  contact->heloname=rou_freestr(contact->heloname);
  contact->locname=rou_freestr(contact->locname);
  contact->info=rou_freestr(contact->info);
  contact->attime=rou_freestr(contact->attime);
  contact->mentunid=rou_freestr(contact->mentunid);
  contact->listenip=rou_freestr(contact->listenip);
  if (contact->dbgname!=(char *)0) {
    (void) fclose(stdout);
    if ((debug<3)&&(contact->bigtrouble==false))
      (void) unlink(contact->dbgname);
    (void) free(contact->dbgname);
    }
  debug=contact->prvdbg;
  contact->session=rou_freestr(contact->session);
  (void) free(contact);
  contact=(CTTYP *)0;
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle exchange with remote*/
/*	SMTP client, if the command can't be	*/
/*	handle in this routine, information is	*/
/*	sent back to the calling routine.	*/
/*						*/
/************************************************/
CTTYP *eml_domails(CTTYP *cta)

{
#define	OPDM	"lvleml.c:eml_domails,"

const VOCTYP *kwlst;
CNFTYP *cnf;
int proceed;

cta->code=s_unknown;
if ((cnf=cnf_readconfig())==(CNFTYP *)0) {
  (void) alert(0,"modtrp.c:gotrequest, Unable to load configuration!");
  return cta;
  }
kwlst=getkwlst(cta->proto);
proceed=true;
while (proceed==true) {
  VOCTYP *voc;
  ACTTYP *acting;
  char *buffer;
 
  if ((buffer=tls_waitforaction(MXTIME,cta->tls))==(char *)0) {
    cta->code=s_timeout;
    break;
    }
  (void) que_cleanstring(buffer);
  cta->code=s_timeout;
  voc=rou_getvoca(kwlst,buffer);
  if ((voc==(VOCTYP *)0)||(voc->code==c_unknown)) {
    (void) eml_printfcrlf(cta->tls,eml_ascstat(SYNERR,buffer));
    break;
    }
  if ((acting=(ACTTYP *)voc->ptr)==(ACTTYP *)0) {
    (void) alert(0,"%s VOC->%d, without action (bug!)",OPDM,voc->code);
    break;
    }
  switch (voc->code) {
    case c_auth	:
    case c_chkid:	/*DO NOT standardize		*/
      break;
    default	:
      (void) que_standardize(buffer);
      break;
    }
  cta->code=acting->scode;
  proceed=acting->proceed;
  switch (voc->code) {
    case c_help		:
      (void) dohelp(cta->tls,kwlst,buffer);
      break;
    case c_data		:
      if (cta->frm==(FRMTYP *)0) {
        (void) eml_printfcrlf(cta->tls,"%d Need MAIL command",BADSEQ);
        proceed=true;
        break;  /*looping again */
        }
      if (countrcpts(cta->rcpts)==0) {
        (void) eml_printfcrlf(cta->tls,"%d Need RCPT (recipient)",BADSEQ);
        proceed=true;
        break;  /*looping again */
        }
      break;
    case c_noop	:
      (void) eml_printfcrlf(cta->tls,"%d OK",CMDOK);
      break;
    case c_reset	:
      break;
    case c_tls	:
      (void) eml_printfcrlf(cta->tls,"%d Ready to start TLS",SIGNON);
      (void) tls_gotls(cta->tls,true,cnf->SSLciphrs);
      cta->cipherid=tls_getcipher(cta->tls);
      break;
    case c_auth	:
      proceed=checkviaauth(cnf->TLSauth,cta,buffer);
      break;
    case c_mail		:
      if (eml_checkfrom(cta,buffer)==true) {
	cta->start=rou_systime();
	cta->toobig=false;
        (void) journalling(false,EVJRL,"%s: %s",MENID,cta->session);
        (void) journalling(false,EVJRL,"%s: [%s]/<%s>",
				       CLIP,cta->ipremote,cta->peername);
        (void) journalling(false,EVJRL,"%s: <%s>",MFROM,cta->frm->from);
	(void) eml_printfcrlf(cta->tls,"%d <%s>... sender ok",
					CMDOK,cta->frm->from);
	}
      break;
    case c_chkid	:
      cta->info=rou_freestr(cta->info);
      if ((cta->info=checkid(cta,buffer))==(char *)0)
        proceed=true;
      break;
    case c_ckid		:
      cta->info=rou_freestr(cta->info);
      cta->info=strdup(buffer);
      break;
    case c_goodmail	:
      cta->info=rou_freestr(cta->info);
      if (cta->dfltcodeok==true)
        cta->info=strdup(buffer);
      else {
        (void) eml_printfcrlf(cta->tls,eml_ascstat(BADAUTH,"dfltcode"));
	proceed=true;
	}	
      break;
    case c_copy	:
      cta->info=rou_freestr(cta->info);
      if (cta->dfltcodeok==true)
        (void) eml_printfcrlf(cta->tls,que_copy(buffer));
      else {
        (void) eml_printfcrlf(cta->tls,eml_ascstat(BADAUTH,"dfltcode"));
	proceed=true;
	}
      break;
    case c_rcpt		:
      proceed=true;
      cta->info=rou_freestr(cta->info);
      if ((cta->info=eml_checkto(cta,buffer))!=(char *)0)
	proceed=false;
      break;
    case c_ehlo		:
      if (eml_checkhelo(cta,buffer,'-')==true) {
        (void) eml_printfcrlf(cta->tls,"%d-ENHANCEDSTATUSCODES",CMDOK);
        (void) eml_printfcrlf(cta->tls,"%d-8BITMIME",CMDOK);
        (void) eml_printfcrlf(cta->tls,"%d-SIZE",CMDOK);
	if ((cnf->TLSauth==false)||(cta->cipherid!=(char *)0))
          (void) eml_printfcrlf(cta->tls,"%d-AUTH PLAIN LOGIN",CMDOK);
	if (cta->cipherid==(char *)0)
          (void) eml_printfcrlf(cta->tls,"%d-STARTTLS",CMDOK);
        (void) eml_printfcrlf(cta->tls,"%d HELP",CMDOK);
	}
      break;
    case c_helo		:
      (void) eml_checkhelo(cta,buffer,' ');
      break;
    case c_quit		:	/*nothing to do	anymore	*/
      break;			/*getting out		*/
    case c_orgn		:
      if (debug>0)
        (void) doorgn(cta,buffer);
      else 
        (void) eml_printfcrlf(cta->tls,"%d Command forbiden",SYNERR);
      break;
    default		:
      (void) alert(0,"%s VOC->%d, can't reach (bug!)",OPDM,voc->code);
      break;
    }
  buffer=rou_freestr(buffer);
  if (rou_checkprocess(getppid())==false) {
    cta->code=s_timeout;
    quickout=true;
    proceed=false;
    }
  }
cnf=cnf_freeconfig(cnf);
return cta;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a TCP connexion to a	*/
/*	SMTP daemon on a specific channel	*/
/*						*/
/************************************************/
CTTYP *eml_openmails(char *sendmail,int port,int swait)

{
#define OPSN	"lvleml.c:eml_openmails,"

CTTYP *cta;
TLSFILE *tls;

cta=(CTTYP *)0;
if ((tls=tls_opentcp(sendmail,port,swait))!=(TLSFILE *)0) {
  int proceed;
  int phase;

  phase=1;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*trouble, trouble		*/
        (void) alert(0,"%s Trouble to open channel to '%s'",OPSN,sendmail);
	cta=eml_itsover(cta);
        (void) tls_close(tls);
	proceed=false;
        break;
      case 1	: {	/*Lets get first prompt		*/
        char *reply;

        reply=(char *)0;
	if (getreply(tls,&reply)!=SIGNON) {
          (void) alert(0,"%s Invalid prompt from '%s' (prompt=<%s>)",
			  OPSN,sendmail,reply);
	  phase=-1;
	  }
	reply=rou_freestr(reply);
        }
        break;
      case 2	:	/*Lets prepare contact info	*/
        cta=(CTTYP *)calloc(1,sizeof(CTTYP));
        cta->tls=tls;
	cta->session=strdup(appname);
        cta->locname=tls_getaddrname(tls,pr_locname,"NO_HOST_NAME");
        cta->peername=tls_getaddrname(tls,pr_remname,NORVERS);
        cta->ipremote=tls_getaddrname(tls,pr_remip,"?.?.?.?");
        cta->hostname=strdup(sendmail);
        cta->remoteport=port;
        break;
      case 3	:	/*doing Hello			*/
	if (doehlo(cta)==false) {
          (void) alert(0,"%s (pid=%05d) Unable to carry EHLO with '%s'",
			 OPSN,getpid(),sendmail);
	  phase=-1;
	  }
	break;
      case 4	:	/*if remote system is TLS, lets go TLS	*/
	if ((cta->capable&CPTLS)!=0) 
	  (void) establishTLS(cta,sendmail);
        break;
      default	:	/*everything well, lets go		*/
	proceed=false;
        break;
      }
    phase++;
    }
  }
return cta;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a string to remote	*/
/*	SMTP channel but add an CRLF at the	*/
/*	of the line.				*/
/*						*/
/************************************************/
void eml_printfcrlf(TLSFILE *tls,const char *fmt,...)

{
va_list args;
char *local;

va_start(args,fmt);
local=rou_dynavsprintf(fmt,args);
local=(char *)realloc(local,strlen(local)+10);
(void) strcat(local,"\r\n");
(void) tls_write(tls,local,strlen(local));
(void) free(local);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage CMD exhange with	*/
/*	a remote SMTP server.			*/
/*						*/
/************************************************/
int eml_sendcmd(CTTYP *contact,int expected,char *fmt,...)

{
#define CMDMRK	"lvleml.c:eml_sendcmd, "
#define	WDATA	60	/*wait 60 second for data	*/

int status;
va_list args;
char *local;

status=ERRPROC;
va_start(args,fmt);
local=rou_dynavsprintf(fmt,args);
local=(char *)realloc(local,strlen(local)+10);
(void) tracing(TRALOG,10,"%05d sending <%s>",getpid(),local);
(void) strcat(local,"\r\n");
if (tls_write(contact->tls,local,strlen(local))>0) {
  contact->info=rou_freestr(contact->info);
  status=getreply(contact->tls,&(contact->info));
  switch (status) {
    case ERRPROC	:
      if (tls_iseof(contact->tls)==false) 
        (void) tracing(TRALOG,0,"%05d %s no answer from smtp "
				"server <%s/%s:%d>",getpid(),CMDMRK,
			         contact->peername,contact->ipremote,
				 contact->remoteport);
      contact->info=strdup(eml_ascstat(status));
      break;
    default		:
      contact->lastxchg=rou_systime();
      if (status!=expected) {
        (void) tracing(TRALOG,5,"%05d %s Unexpected answer <%s> to "
				"command <%s>",getpid(),CMDMRK,
				 contact->info,local);
	}
      break;
    }
  }
(void) free(local);
va_end(args);
return status; 
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a message in the boxe	*/
/*	IPC.					*/
/*	Procedure is waiting for remote to	*/
/*	answer and suppose there is only one	*/
/*	message from destination back at any	*/
/*	given time.				*/
/*						*/
/************************************************/
char *eml_xchgmsg(BOXTYP *boxe,pid_t dest,char *fmt,...)

{
#define	OPXC	"lvleml.c:eml_xchgmsg,"

#define	MXWMSG	120	/*wait 120 sec		*/
#define	MXMSG	5000	/*max message size	*/

char *got;
va_list args;
char *msgref;
char *msg;

got=(char *)0;
va_start(args,fmt);
msg=(char *)calloc(MXMSG,sizeof(char));
(void) vsnprintf(msg,MXMSG-1,fmt,args);
if ((msgref=ipc_postmsg(boxe,dest,msg))!=(char *)0) {
  int i;

  for (i=0;i<MXWMSG;i++) {
    if ((got=ipc_getreply(boxe,msgref))!=(char *)0) 
      break;
    (void) usleep(1000000);
    }
  msgref=rou_freestr(msgref);
  }
if (got==(char *)0)
  (void) alert(0,"%s no answer from '%d'",OPXC,dest);
(void) free(msg);
va_end(args);
return got;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check is SMTP connexion is	*/
/*	still active.				*/
/*						*/
/************************************************/
int eml_smtpidle(CHKTYP *smtp,int idle)

{
int ok;

ok=false;
if (smtp->contact!=(CTTYP *)0) {
  ok=(tls_idle(smtp->contact->tls)>idle);
  }
return ok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect and store E-mail	*/
/*	data, received from remote site.	*/
/*						*/
/************************************************/
int eml_collectdata(CTTYP *cta)

{
#define	HDRLIM	32768	/*max header size	*/
#define	DTWAIT	300	/*wait at least 300 	*/
			/*seconds for data input*/

int status;
FILE *store;
char ext[10];
char filename[300];

status=false;
(void) snprintf(ext,sizeof(ext),".%02d-D",cta->sequence);
(void) rou_mkname(filename,sizeof(filename),d_spool,DATADIR,cta->session,ext);
cta->dataname=strdup(filename);
cta->mtaille=(u_long)0;
if ((store=fopen(cta->dataname,"w"))!=(FILE *)0) {
  enum {
       mm_header,	/*in header mode	*/
       mm_data,		/*in data mode		*/
       mm_dummy		/*do not care!		*/
       }inmode;
  int proceed;

  (void) mkreceived(store,cta);
  (void) eml_printfcrlf(cta->tls,eml_ascstat(DATAOK));
  inmode=mm_header;
  proceed=true;
  while (proceed==true) {
    char **data;

    if ((data=getstream(cta))==(char **)0)
      break;
    else {
      int d;

      for (d=0;data[d]!=(char *)0;d++) {
        register char *buffer;

        buffer=data[d];
        cta->mtaille+=strlen(buffer);
        (void) que_cleanstring(buffer); 
        if ((buffer[0]=='.')&&(buffer[1]==0)) {
          status=true;
	  proceed=false;
          break;			/*reach end of transmission	*/
          }
	if (strlen(buffer)==2) {	/*cas du ".' seul sur une ligne	*/
	  if ((buffer[0]=='.')&&(buffer[1]=='.'))
	    buffer[1]='\000';
	  }
        switch (inmode) {
          case mm_header	:
             if (buffer[0]!='\000') {
	      cta->headers=(char **)rou_addveclst((void **)cta->headers,
					          (void *)strdup(buffer));
	      if (cta->mtaille>HDRLIM) {
	        char info[100];
    
                (void) snprintf(info,sizeof(info),"Mail header too large "
						  "(max %d)",HDRLIM);
                (void) eml_printfcrlf(cta->tls,eml_ascstat(MALABRT,info));
	        if (cta->rcpts!=(RCPTYP **)0) {
	          int i;
    
	          for (i=0;cta->rcpts[i]!=(RCPTYP *)0;i++)
                    (void) journalling(false,EVJRL,"RCPT: <%s>",
						    cta->rcpts[i]->rcpt);
	          }
                (void) journalling(false,EVJRL,"Abort! E-mail header too big");
	        cta->headers=rou_freestrlst(cta->headers);
      	        cta->code=s_timeout;	/*lest say we have a time out	*/
	        inmode=mm_dummy;	/*and wait for end of E-mail	*/
                }
	      }
            else {
              (void) fprintf(store,"%sVersion: %s\n",XCLEM,rou_getvers());
              (void) fprintf(store,"%sID: <%s>\n",XCLEM,cta->session);
              (void) fprintf(store,"%sVirus-checker: %s\n",XCLEM,cta->clamvers);
              inmode=mm_data;
	      }
			    /*No break, need to store info	*/
          case mm_data	:
            (void) fprintf(store,"%s\n",buffer);
	    break;
          case mm_dummy	:
	    break;
          }
	}
      data=rou_freestrlst(data);
      }
    }
  (void) fclose(store);
  }
else {
  (void) eml_printfcrlf(cta->tls,eml_ascstat(ERRPROC,"Space Missing"));
  (void) core_dump("lvleml.c:collectdata, Unable to open <%s> (error=<%s>)",
		    cta->dataname,strerror(errno));
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send data information over	*/
/*	an SMTP channel. If not successfull	*/
/*	return False.				*/
/*						*/
/************************************************/
int xeml_senddata(CTTYP *cta,char *title,char **data)

{
#define	OPSE	"lvleml.c:eml_senddata,"

int status;
int proceed;
int phase;

status=true;
phase=0;
proceed=true;
while (proceed==true) {
  int chkerr;

  chkerr=0;
  switch (phase) {
    case   0	:
      if ((chkerr=eml_sendcmd(cta,DATAOK,title))!=DATAOK){
	(void) alert(0,"%s remote site denied report (error='%d')",OPSE,chkerr);
	proceed=false;
	}
      break;
    case   1	:	/*sending the report contents	*/
      if (data!=(char **)0) {
	int i;

	for (i=0;data[i]!=(char *)0;i++) {
          (void) eml_printfcrlf(cta->tls,"%s",data[i]);
	  }
        }
      break;
    case   2	:	/*sending end of data		*/
      (void) eml_printfcrlf(cta->tls,".");
      break;
    case   3	:	/*collecting all status		*/
      cta->info=rou_freestr(cta->info);
      while (true) {
	int acting;
	int pos;
	char *respnd;

        if ((respnd=tls_waitforaction(MXTIME,cta->tls))==(char *)0) {
	  (void) alert(0,"%s remote site timeout",OPSE);
	  status=false;
	  break;
	  }
        (void) que_cleanstring(respnd); 
        cta->info=rou_memstrcat(cta->info,"%s\n",respnd);	
	if (sscanf(respnd,"%d%n",&acting,&pos)==1) {
	  if (respnd[pos]!='-') {
	    if (acting!=CMDOK)
	      status=false;
	    break;
	    }
	  }
	respnd=rou_freestr(respnd);
	}
      break;
    default	:	/*everything went fine		*/
      proceed=false;
      break;	
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a data stream		*/
/*	information over an SMTP channel.	*/
/*	stream information do not require a	*/
/*	remote acknowledge...			*/
/*	Answer to request can be receive way	*/
/*	later.					*/
/*						*/
/************************************************/
int eml_sendstream(CTTYP *cta,char *title,char *seq,char *msg)

{
int status;
int proceed;
int phase;


status=true;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:
      (void) eml_printfcrlf(cta->tls,"%s %s",title,seq);
      break;
    case   1	:	/*sending the stream itself	*/
      if (msg!=(char *)0) {
	char **split;
	int num;
	int i;

	split=eml_splitmsg(msg);
	num=rou_vecsize((void **)split);	
	for (i=0;i<num;i++)
	  (void) eml_printfcrlf(cta->tls,"%s",split[i]);
	split=rou_freestrlst(split);
	}
      break;
    case   2	:	/*sending the end of stream	*/
      (void) eml_printfcrlf(cta->tls,".",title,seq);
      break;
    default	:	/*safe guard			*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open checker or sender	*/
/*	auxiliary channel according config 	*/
/*	informations.				*/
/*						*/
/************************************************/
CHKTYP *eml_openauxiliary(SKRTYP *skr,char *uniqid)

{
CHKTYP *checker;

checker=(CHKTYP *)0;
if (strcasecmp(skr->hostname,KEEPLOC)==0) {
  checker=(CHKTYP *)calloc(1,sizeof(CHKTYP));
  checker->control=c_local;
  checker->hostname=strdup(skr->hostname);
  }
else {
  if ((checker=openremotechecker(skr,uniqid))!=(CHKTYP *)0) {
    checker->relay=skr->relay;
    checker->control=c_standard;
    if ((checker->contact->capable&CPRQT)!=0) 
      checker->control=c_mentor;
    }
  }
if (checker!=(CHKTYP *)0) {
  if (skr->dfs!=(DFSTYP **)0) {
    int i;

    for (i=0;skr->dfs[i]!=(DFSTYP *)0;i++) {
      DFSTYP *df;
      DFSTYP *nd;

      df=skr->dfs[i];
      nd=eml_mkdf(df->domain,df->webgood,df->spamlim,df->qualoc,df->quahead,df->quaextra);
      checker->dfs=(DFSTYP **)rou_addveclst((void **)checker->dfs,(void *)nd);
      }
    }
  }
(void) journalling(true,APJRL,"");
return checker;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close auxiliary channel	*/
/*						*/
/************************************************/
CHKTYP *eml_closeauxiliary(CHKTYP *checker)

{
if (checker!=(CHKTYP *)0) {
  checker->dfs=eml_freedfslst(checker->dfs);
  checker->hostname=rou_freestr(checker->hostname);
  switch(checker->control) {
    case (c_local)	:/*local checker down	*/
      (void) free(checker);
      checker=(CHKTYP *)0;
      break;
    case (c_mentor)	:
      (void) senddropme(checker);
				/*NO BREAK	*/
    case (c_standard)	:
      if (debug>0) 
        (void) journalling(true,APJRL,"Disconnecting from Checker <%s,%d>",
			        checker->hostname,
				checker->contact->remoteport);
      checker=eml_closesmtp(checker);
      break;
    }
  }
return checker;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a file (E-mail) to	*/
/*	an auxiliary server (could be local)	*/
/*						*/
/************************************************/
int eml_sendingauxiliary(CHKTYP *auxi,char *queudir,char *emailid)

{
#define	OPCY	"lvleml.c:eml_sendingauxiliary"

int status;

status=false;
if (auxi!=(CHKTYP *)0) {
  status=true;
  switch (auxi->control) {
    case c_local	:
    case c_mentor	:
      break;
    case c_standard	:
      if (eml_sendcmd(auxi->contact,CMDOK,RSET)!=CMDOK) {
        (void) alert(0,"%s Lost contact with SMTP <%s> responding!",
			OPCY,auxi->hostname);
	status=false;
	}
      break;
    }
  if (status==true)
    status=dispatching(auxi,queudir,emailid);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle the unique session	*/
/*	ID, this ID is a cross over from	*/
/*	the current session unique ID and	*/
/*	from a Mentor checker if available	*/
/*						*/
/************************************************/
void eml_setsession(TLSFILE *tls,CTTYP *cta)

{
char *locip;
char *uniname;
char session[100];

locip=tls_getaddrname(tls,pr_locip,"?.?.?.?");
uniname=rou_uniquename();
(void) snprintf(session,sizeof(session),"%s-%s",locip,uniname);
(void) free(uniname);
(void) free(locip);
cta->session=rou_freestr(cta->session);
cta->session=strdup(session);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find out about sender and	*/
/*	checker for a specific domain.		*/
/*						*/
/************************************************/
DOMTYP *eml_findskr(SKRTYP **skrlst,char *domain,int relais)

{
DOMTYP *dominfo;
int phase;
int proceed;

dominfo=(DOMTYP *)calloc(1,sizeof(DOMTYP));
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*find a sender already open	*/
      dominfo->sndr1=locremote(skrlst,domain,p_sender);
      if (dominfo->sndr1==(SKRTYP *)0)
	phase++;    
      break;
    case 1	:	/*find a checker already open	*/
      dominfo->chkr1=locremote(skrlst,domain,p_checker);
      proceed=false;	/*sender found may be no checker*/
      break;
    case 2	:	/*no domain, lets try default	*/
      dominfo->sndr1=locremote(skrlst,DFLTDOM,p_sender);
      break;
    case 3	:	/*lets try default for checker	*/
      dominfo->chkr1=locremote(skrlst,DFLTDOM,p_checker);
      break;
    case 4	:	/*are we in relay mode		*/
      proceed=relais;
      break;
    case 5	:	/*no domain, lets try relais	*/
      dominfo->sndr2=locremote(skrlst,RLAYDOM,p_sender);
      break;
    default	:	/*SAFE guard			*/
      proceed=false;
      break;
    }
  phase++;
  }
if ((dominfo->sndr1==(SKRTYP *)0)&& (dominfo->sndr2==(SKRTYP *)0)) {
  (void) free(dominfo);
  dominfo=(DOMTYP *)0;
  }
return dominfo;
}
